import enb.plugins


class FileVersionExample(enb.plugins.Plugin):
    """Template and usage example for the FileVersionTable class.
    """
    name = "file_version_example"
    author = ["Miguel Hernández-Cabronero"]
    label = "Template and usage example for the FileVersionTable class for dataset curation"
    tags = {"project", "documentation"}
    tested_on = {"linux"}
